@@ -1,7 +1,7 @@
 /*
  * Copyright 2003-2006 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
- * $Header: /var/cvsroot/gentoo-projects/pax-utils/paxelf.c,v 1.34 2006/01/10 01:40:15 vapier Exp $
+ * $Header: /var/cvsroot/gentoo-projects/pax-utils/paxelf.c,v 1.42 2006/02/05 17:10:52 solar Exp $
  *
  * Copyright 2005-2006 Ned Ludd        - <solar@gentoo.org>
  * Copyright 2005-2006 Mike Frysinger  - <vapier@gentoo.org>
@@ -92,16 +92,29 @@ static pairtype elf_etypes[] = {
 	QUERY(ET_HIPROC),
 	{ 0, 0 }
 };
-const char *get_elfetype(elfobj *elf)
+
+int get_etype(elfobj *elf)
 {
 	int type;
 	if (elf->elf_class == ELFCLASS32)
 		type = EGET(EHDR32(elf->ehdr)->e_type);
 	else
 		type = EGET(EHDR64(elf->ehdr)->e_type);
-	return find_pairtype(elf_etypes, type);
+        return type;
 }
 
+const char *get_elfetype(elfobj *elf)
+{
+	return find_pairtype(elf_etypes, get_etype(elf));
+}
+
+void print_etypes(FILE *stream)
+{
+        int i;
+	for (i = 0; elf_etypes[i].str; ++i) {
+		fprintf(stream, " (%d) = %s\n", elf_etypes[i].value, elf_etypes[i].str);
+	}
+}
 /* translate elf EM_ defines */
 static pairtype elf_emtypes[] = {
 	QUERY(EM_NONE),
@@ -197,6 +210,7 @@ static pairtype elf_ptypes[] = {
 	QUERY(PT_SHLIB),
 	QUERY(PT_PHDR),
 	QUERY(PT_TLS),
+	QUERY(PT_NUM),
 	QUERY(PT_GNU_EH_FRAME),
 	QUERY(PT_GNU_STACK),
 	QUERY(PT_GNU_RELRO),
@@ -326,45 +340,40 @@ const char *get_elfstttype(int type)
 	((buff[EI_CLASS] == ELFCLASS32 || buff[EI_CLASS] == ELFCLASS64) && \
 	 (buff[EI_DATA] == ELFDATA2LSB || buff[EI_DATA] == ELFDATA2MSB) && \
 	 (buff[EI_VERSION] == EV_CURRENT))
-elfobj *_readelf(const char *filename, int read_only)
+elfobj *readelf_buffer(const char *filename, char *buffer, size_t buffer_len)
 {
-	struct stat st;
-	int fd;
 	elfobj *elf;
 
-	if (stat(filename, &st) == -1)
-		return NULL;
-
-	if ((fd = open(filename, (read_only ? O_RDONLY : O_RDWR))) == -1)
-		return NULL;
-
 	/* make sure we have enough bytes to scan e_ident */
-	if (st.st_size <= EI_NIDENT)
-		goto close_fd_and_return;
+	if (buffer == NULL || buffer_len < EI_NIDENT)
+		return NULL;
 
 	elf = (elfobj*)malloc(sizeof(*elf));
 	if (elf == NULL)
-		goto close_fd_and_return;
+		return NULL;
 	memset(elf, 0x00, sizeof(*elf));
 
-	elf->fd = fd;
-	elf->len = st.st_size;
-	elf->data = (char*)mmap(0, elf->len, PROT_READ | (read_only ? 0 : PROT_WRITE), (read_only ? MAP_PRIVATE : MAP_SHARED), fd, 0);
-	if (elf->data == (char*)MAP_FAILED) {
-		warn("mmap on '%s' of %li bytes failed :(", filename, (unsigned long)elf->len);
-		goto free_elf_and_return;
+	elf->fd = -1;
+	elf->len = buffer_len;
+	elf->data = buffer;
+	elf->data_end = buffer + buffer_len;
+
+	/* make sure we have an elf */
+	if (!IS_ELF_BUFFER(elf->data)) {
+free_elf_and_return:
+		free(elf);
+		return NULL;
 	}
 
-	if (!IS_ELF_BUFFER(elf->data)) /* make sure we have an elf */
-		goto unmap_data_and_return;
-	if (!DO_WE_LIKE_ELF(elf->data)) { /* check class and stuff */
+	/* check class and stuff */
+	if (!DO_WE_LIKE_ELF(elf->data)) {
 		warn("we no likey %s: {%s,%s,%s,%s}",
 		     filename,
 		     get_elfeitype(EI_CLASS, elf->data[EI_CLASS]),
 		     get_elfeitype(EI_DATA, elf->data[EI_DATA]),
 		     get_elfeitype(EI_VERSION, elf->data[EI_VERSION]),
 		     get_elfeitype(EI_OSABI, elf->data[EI_OSABI]));
-		goto unmap_data_and_return;
+		goto free_elf_and_return;
 	}
 
 	elf->filename = filename;
@@ -379,33 +388,44 @@ elfobj *_readelf(const char *filename, int read_only)
 
 #define READELF_HEADER(B) \
 	if (elf->elf_class == ELFCLASS ## B) { \
+		char invalid; \
 		Elf ## B ## _Ehdr *ehdr = EHDR ## B (elf->ehdr); \
 		Elf ## B ## _Off size; \
 		/* verify program header */ \
-		if (EGET(ehdr->e_phnum) > 0) { \
+		invalid = 0; \
+		if (EGET(ehdr->e_phnum) <= 0) \
+			invalid = 1; /* this is not abnormal so dont warn */ \
+		else if (EGET(ehdr->e_phentsize) != sizeof(Elf ## B ## _Phdr)) \
+			invalid = 3; \
+		else { \
 			elf->phdr = elf->data + EGET(ehdr->e_phoff); \
 			size = EGET(ehdr->e_phnum) * EGET(ehdr->e_phentsize); \
 			if (elf->phdr < elf->ehdr || /* check overflow */ \
 			    elf->phdr + size < elf->phdr || /* before start of mem */ \
 			    elf->phdr + size > elf->ehdr + elf->len) /* before end of mem */ \
-			{ \
-				warn("%s: Invalid program header info", filename); \
-				elf->phdr = NULL; \
-			} \
-		} else \
+				invalid = 2; \
+		} \
+		if (invalid > 1) \
+			warn("%s: Invalid program header info (%i)", filename, invalid); \
+		if (invalid) \
 			elf->phdr = NULL; \
 		/* verify section header */ \
-		if (EGET(ehdr->e_shnum) > 0) { \
+		invalid = 0; \
+		if (EGET(ehdr->e_shnum) <= 0) \
+			invalid = 1; /* this is not abnormal so dont warn */ \
+		else if (EGET(ehdr->e_shentsize) != sizeof(Elf ## B ## _Shdr)) \
+			invalid = 3; \
+		else { \
 			elf->shdr = elf->data + EGET(ehdr->e_shoff); \
 			size = EGET(ehdr->e_shnum) * EGET(ehdr->e_shentsize); \
 			if (elf->shdr < elf->ehdr || /* check overflow */ \
 			    elf->shdr + size < elf->shdr || /* before start of mem */ \
 			    elf->shdr + size > elf->ehdr + elf->len) /* before end of mem */ \
-			{ \
-				warn("%s: Invalid section header info", filename); \
-				elf->shdr = NULL; \
-			} \
-		} else \
+				invalid = 2; \
+		} \
+		if (invalid > 1) \
+			warn("%s: Invalid section header info (%i)", filename, invalid); \
+		if (invalid) \
 			elf->shdr = NULL; \
 	}
 	READELF_HEADER(32)
@@ -413,22 +433,68 @@ elfobj *_readelf(const char *filename, int read_only)
 	/* { char *p; strncpy(elf->basename, (p = strrchr(filename, '/')) == NULL ? "?" : p+1 , sizeof(elf->basename)); } */
 
 	return elf;
-
-unmap_data_and_return:
-	munmap(elf->data, elf->len);
-free_elf_and_return:
-	free(elf);
-close_fd_and_return:
-	close(fd);
-	return NULL;
 }
-elfobj *readelf(const char *filename) { return _readelf(filename, 1); }
+elfobj *_readelf_fd(const char *filename, int fd, size_t len, int read_only)
+{
+	char *buffer;
+	elfobj *ret;
+
+	if (len == 0) {
+		struct stat st;
+		if (fstat(fd, &st) == -1)
+			return NULL;
+		len = st.st_size;
+		if (len == 0)
+			return NULL;
+	}
+
+	buffer = (char*)mmap(0, len, PROT_READ | (read_only ? 0 : PROT_WRITE), (read_only ? MAP_PRIVATE : MAP_SHARED), fd, 0);
+	if (buffer == (char*)MAP_FAILED) {
+		warn("mmap on '%s' of %li bytes failed :(", filename, (unsigned long)len);
+		return NULL;
+	}
+
+	ret = readelf_buffer(filename, buffer, len);
+	if (ret == NULL)
+		munmap(buffer, len);
+	else {
+		ret->fd = fd;
+		ret->is_mmap = 1;
+	}
+
+	return ret;
+}
+elfobj *_readelf(const char *filename, int read_only)
+{
+	elfobj *ret;
+	struct stat st;
+	int fd;
+
+	if (stat(filename, &st) == -1)
+		return NULL;
+
+	if ((fd = open(filename, (read_only ? O_RDONLY : O_RDWR))) == -1)
+		return NULL;
+
+	/* make sure we have enough bytes to scan e_ident */
+	if (st.st_size <= EI_NIDENT) {
+close_fd_and_return:
+		close(fd);
+		return NULL;
+	}
+
+	ret = readelf_fd(filename, fd, st.st_size);
+	if (ret == NULL)
+		goto close_fd_and_return;
+
+	return ret;
+}
 
 /* undo the readelf() stuff */
 void unreadelf(elfobj *elf)
 {
-	munmap(elf->data, elf->len);
-	close(elf->fd);
+	if (elf->is_mmap) munmap(elf->data, elf->len);
+	if (elf->fd != -1) close(elf->fd);
 	free(elf);
 }
 
@@ -531,6 +597,7 @@ int elf_max_pt_load(elfobj *elf)
 	Elf ## B ## _Ehdr *ehdr = EHDR ## B (elf->ehdr); \
 	switch (EGET(ehdr->e_ident[EI_OSABI])) { \
 	case ELFOSABI_NONE: \
+	case ELFOSABI_LINUX: \
 	case ELFOSABI_ARM:     return 2; \
 	case ELFOSABI_OPENBSD: return 7; \
 	} }
