@@ -1,7 +1,7 @@
 /*
  * Copyright 2005-2006 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
- * $Header: /var/cvsroot/gentoo-projects/pax-utils/paxelf.h,v 1.39 2006/01/10 01:40:15 vapier Exp $
+ * $Header: /var/cvsroot/gentoo-projects/pax-utils/paxelf.h,v 1.43 2006/02/05 17:10:52 solar Exp $
  *
  * Copyright 2005-2006 Ned Ludd        - <solar@gentoo.org>
  * Copyright 2005-2006 Mike Frysinger  - <vapier@gentoo.org>
@@ -16,10 +16,11 @@ typedef struct {
 	void *ehdr;
 	void *phdr;
 	void *shdr;
-	char *data;
+	char *data, *data_end;
 	char elf_class;
 	off_t len;
 	int fd;
+	int is_mmap;
 	const char *filename;
 	const char *base_filename;
 } elfobj;
@@ -42,8 +43,11 @@ typedef struct {
 extern char *pax_short_hf_flags(unsigned long flags);
 extern char *pax_short_pf_flags(unsigned long flags);
 extern char *gnu_short_stack_flags(unsigned long flags);
+extern elfobj *readelf_buffer(const char *filename, char *buffer, size_t buffer_len);
+extern elfobj *_readelf_fd(const char *filename, int fd, size_t len, int read_only);
+#define readelf_fd(filename, fd, len) _readelf_fd(filename, fd, len, 1)
 extern elfobj *_readelf(const char *filename, int read_only);
-extern elfobj *readelf(const char *filename);
+#define readelf(filename) _readelf(filename, 1)
 extern void unreadelf(elfobj *elf);
 extern const char *get_elfeitype(int ei_type, int type);
 extern const char *get_elfetype(elfobj *elf);
@@ -54,6 +58,8 @@ extern const char *get_elfshttype(int type);
 extern const char *get_elfstttype(int type);
 extern void *elf_findsecbyname(elfobj *elf, const char *name);
 extern int elf_max_pt_load(elfobj *elf);
+extern int get_etype(elfobj *elf);
+extern void print_etypes(FILE *);
 
 /* PaX flags (to be read in elfhdr.e_flags) */
 #define HF_PAX_PAGEEXEC      1   /* 0: Paging based non-exec pages */
