@@ -1,7 +1,7 @@
 /*
  * Copyright 2003-2006 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
- * $Header: /var/cvsroot/gentoo-projects/pax-utils/scanelf.c,v 1.103 2006/01/11 23:46:21 vapier Exp $
+ * $Header: /var/cvsroot/gentoo-projects/pax-utils/scanelf.c,v 1.121 2006/02/05 17:10:52 solar Exp $
  *
  * Copyright 2003-2006 Ned Ludd        - <solar@gentoo.org>
  * Copyright 2004-2006 Mike Frysinger  - <vapier@gentoo.org>
@@ -9,7 +9,7 @@
 
 #include "paxinc.h"
 
-static const char *rcsid = "$Id: scanelf.c,v 1.103 2006/01/11 23:46:21 vapier Exp $";
+static const char *rcsid = "$Id: scanelf.c,v 1.121 2006/02/05 17:10:52 solar Exp $";
 #define argv0 "scanelf"
 
 #define IS_MODIFIER(c) (c == '%' || c == '#')
@@ -17,10 +17,13 @@ static const char *rcsid = "$Id: scanelf.c,v 1.103 2006/01/11 23:46:21 vapier Ex
 
 
 /* prototypes */
+static int scanelf_elfobj(elfobj *elf);
+static int scanelf_elf(const char *filename, int fd, size_t len);
+static int scanelf_archive(const char *filename, int fd, size_t len);
 static void scanelf_file(const char *filename);
 static void scanelf_dir(const char *path);
-static void scanelf_ldpath();
-static void scanelf_envpath();
+static void scanelf_ldpath(void);
+static void scanelf_envpath(void);
 static void usage(int status);
 static void parseargs(int argc, char *argv[]);
 static char *xstrdup(const char *s);
@@ -30,10 +33,12 @@ static void xstrncat(char **dst, const char *src, size_t *curr_len, size_t n);
 static inline void xchrcat(char **dst, const char append, size_t *curr_len);
 
 /* variables to control behavior */
+static char match_etypes[126] = "";
 static char *ldpaths[256];
 static char scan_ldpath = 0;
 static char scan_envpath = 0;
 static char scan_symlink = 1;
+static char scan_archives = 0;
 static char dir_recurse = 0;
 static char dir_crossmount = 1;
 static char show_pax = 0;
@@ -57,7 +62,7 @@ static char fix_elf = 0;
 static char gmatch = 0;
 static char use_ldcache = 0;
 
-
+int match_bits = 0;
 caddr_t ldcache = 0;
 size_t ldcache_size = 0;
 
@@ -150,6 +155,7 @@ static char *scanelf_file_phdr(elfobj *elf, char *found_phdr, char *found_relro,
 	multi_stack = multi_relro = multi_load = 0;
 	max_pt_load = elf_max_pt_load(elf);
 
+#define NOTE_GNU_STACK ".note.GNU-stack"
 #define SHOW_PHDR(B) \
 	if (elf->elf_class == ELFCLASS ## B) { \
 	Elf ## B ## _Ehdr *ehdr = EHDR ## B (elf->ehdr); \
@@ -169,7 +175,8 @@ static char *scanelf_file_phdr(elfobj *elf, char *found_phdr, char *found_relro,
 				offset = 4; \
 				check_flags = PF_X; \
 			} else if (EGET(phdr[i].p_type) == PT_LOAD) { \
-				if (multi_load++ > max_pt_load) warnf("%s: more than %i PT_LOAD's !?", elf->filename, max_pt_load); \
+				if (ehdr->e_type == ET_DYN || ehdr->e_type == ET_EXEC) \
+					if (multi_load++ > max_pt_load) warnf("%s: more than %i PT_LOAD's !?", elf->filename, max_pt_load); \
 				found = found_load; \
 				offset = 8; \
 				check_flags = PF_W|PF_X; \
@@ -191,11 +198,16 @@ static char *scanelf_file_phdr(elfobj *elf, char *found_phdr, char *found_relro,
 		/* no program headers which means this is prob an object file */ \
 		Elf ## B ## _Shdr *shdr = SHDR ## B (elf->shdr); \
 		Elf ## B ## _Shdr *strtbl = shdr + EGET(ehdr->e_shstrndx); \
+		char *str; \
+		if ((void*)strtbl > (void*)elf->data_end) \
+			goto skip_this_shdr##B; \
 		check_flags = SHF_WRITE|SHF_EXECINSTR; \
 		for (i = 0; i < EGET(ehdr->e_shnum); ++i) { \
 			if (EGET(shdr[i].sh_type) != SHT_PROGBITS) continue; \
 			offset = EGET(strtbl->sh_offset) + EGET(shdr[i].sh_name); \
-			if (!strcmp((char*)(elf->data + offset), ".note.GNU-stack")) { \
+			str = elf->data + offset; \
+			if (str > elf->data + offset + sizeof(NOTE_GNU_STACK)) continue; \
+			if (!strcmp(str, NOTE_GNU_STACK)) { \
 				if (multi_stack++) warnf("%s: multiple .note.GNU-stack's !?", elf->filename); \
 				flags = EGET(shdr[i].sh_flags); \
 				if (be_quiet && ((flags & check_flags) != check_flags)) \
@@ -209,6 +221,7 @@ static char *scanelf_file_phdr(elfobj *elf, char *found_phdr, char *found_relro,
 				break; \
 			} \
 		} \
+		skip_this_shdr##B: \
 		if (!multi_stack) { \
 			*found_phdr = 1; \
 			shown = 1; \
@@ -322,6 +335,10 @@ static char *scanelf_file_textrels(elfobj *elf, char *found_textrels, char *foun
 				*found_textrels = 1; \
 			/* locate this relocation symbol name */ \
 			sym = SYM ## B (elf->data + EGET(symtab->sh_offset)); \
+			if ((void*)sym > (void*)elf->data_end) { \
+				warn("%s: corrupt ELF symbol", elf->filename); \
+				continue; \
+			} \
 			sym_max = ELF ## B ## _R_SYM(r_info); \
 			if (sym_max * EGET(symtab->sh_entsize) < symtab->sh_size) \
 				sym += sym_max; \
@@ -461,7 +478,7 @@ static void scanelf_file_rpath(elfobj *elf, char *found_rpath, char **ret, size_
 						} \
 					} \
 					if (*r) { \
-						if (fix_elf > 2 || **r == '\0') { \
+						if (fix_elf > 2 || (fix_elf && **r == '\0')) { \
 							/* just nuke it */ \
 							nuke_it##B: \
 							memset(*r, 0x00, offset); \
@@ -553,7 +570,7 @@ static char *lookup_cache_lib(elfobj *elf, char *fname)
 {
 	int fd = 0;
 	char *strs;
-	static char buf[_POSIX_PATH_MAX] = "";
+	static char buf[__PAX_UTILS_PATH_MAX] = "";
 	const char *cachefile = "/etc/ld.so.cache";
 	struct stat st;
 
@@ -584,8 +601,10 @@ static char *lookup_cache_lib(elfobj *elf, char *fname)
 
 		close(fd);
 
-		if (ldcache == (caddr_t)-1)
+		if (ldcache == (caddr_t)-1) {
+			ldcache = 0;
 			return NULL;
+		}
 
 		if (memcmp(((header_t *) ldcache)->magic, LDSO_CACHE_MAGIC, LDSO_CACHE_MAGIC_LEN))
 			return NULL;
@@ -799,11 +818,17 @@ static char *scanelf_file_sym(elfobj *elf, char *found_sym)
 		Elf ## B ## _Shdr *symtab = SHDR ## B (symtab_void); \
 		Elf ## B ## _Shdr *strtab = SHDR ## B (strtab_void); \
 		Elf ## B ## _Sym *sym = SYM ## B (elf->data + EGET(symtab->sh_offset)); \
-		unsigned long cnt = EGET(symtab->sh_size) / EGET(symtab->sh_entsize); \
+		unsigned long cnt = EGET(symtab->sh_entsize); \
 		char *symname; \
+		if (cnt) \
+			cnt = EGET(symtab->sh_size) / cnt; \
 		for (i = 0; i < cnt; ++i) { \
 			if (sym->st_name) { \
 				symname = (char *)(elf->data + EGET(strtab->sh_offset) + EGET(sym->st_name)); \
+				if ((void*)symname > (void*)elf->data_end) { \
+					warnf("%s: corrupt ELF symbols", elf->filename); \
+					continue; \
+				} \
 				if (*find_sym == '*') { \
 					printf("%s(%s) %5lX %15s %s\n", \
 					       ((*found_sym == 0) ? "\n\t" : "\t"), \
@@ -845,50 +870,29 @@ break_out:
 	else
 		return (char *)" - ";
 }
+
+
 /* scan an elf file and show all the fun stuff */
 #define prints(str) write(fileno(stdout), str, strlen(str))
-static void scanelf_file(const char *filename)
+static int scanelf_elfobj(elfobj *elf)
 {
 	unsigned long i;
 	char found_pax, found_phdr, found_relro, found_load, found_textrel, 
 	     found_rpath, found_needed, found_interp, found_bind, found_soname, 
 	     found_sym, found_lib, found_file, found_textrels;
-	elfobj *elf;
-	struct stat st;
 	static char *out_buffer = NULL;
 	static size_t out_len;
 
-	/* make sure 'filename' exists */
-	if (lstat(filename, &st) == -1) {
-		if (be_verbose > 2) printf("%s: does not exist\n", filename);
-		return;
-	}
-	/* always handle regular files and handle symlinked files if no -y */
-	if (S_ISLNK(st.st_mode)) {
-		if (!scan_symlink) return;
-		stat(filename, &st);
-	}
-	if (!S_ISREG(st.st_mode)) {
-		if (be_verbose > 2) printf("%s: skipping non-file\n", filename);
-		return;
-	}
-
 	found_pax = found_phdr = found_relro = found_load = found_textrel = \
 	found_rpath = found_needed = found_interp = found_bind = found_soname = \
 	found_sym = found_lib = found_file = found_textrels = 0;
 
-	/* verify this is real ELF */
-	if ((elf = _readelf(filename, !fix_elf)) == NULL) {
-		if (be_verbose > 2) printf("%s: not an ELF\n", filename);
-		return;
-	}
-
-	if (be_verbose > 1)
-		printf("%s: scanning file {%s,%s}\n", filename,
+	if (be_verbose > 2)
+		printf("%s: scanning file {%s,%s}\n", elf->filename,
 		       get_elfeitype(EI_CLASS, elf->elf_class),
 		       get_elfeitype(EI_DATA, elf->data[EI_DATA]));
-	else if (be_verbose)
-		printf("%s: scanning file\n", filename);
+	else if (be_verbose > 1)
+		printf("%s: scanning file\n", elf->filename);
 
 	/* init output buffer */
 	if (!out_buffer) {
@@ -952,16 +956,16 @@ static void scanelf_file(const char *filename)
 		case 'F':
 			found_file = 1;
 			if (be_wewy_wewy_quiet) break;
-			xstrcat(&out_buffer, filename, &out_len);
+			xstrcat(&out_buffer, elf->filename, &out_len);
 			break;
 		case 'p':
 			found_file = 1;
 			if (be_wewy_wewy_quiet) break;
-			tmp = filename;
+			tmp = elf->filename;
 			if (search_path) {
 				ssize_t len_search = strlen(search_path);
-				ssize_t len_file = strlen(filename);
-				if (!strncmp(filename, search_path, len_search) && \
+				ssize_t len_file = strlen(elf->filename);
+				if (!strncmp(elf->filename, search_path, len_search) && \
 				    len_file > len_search)
 					tmp += len_search;
 				if (*tmp == '/' && search_path[len_search-1] == '/') tmp++;
@@ -971,8 +975,8 @@ static void scanelf_file(const char *filename)
 		case 'f':
 			found_file = 1;
 			if (be_wewy_wewy_quiet) break;
-			tmp = strrchr(filename, '/');
-			tmp = (tmp == NULL ? filename : tmp+1);
+			tmp = strrchr(elf->filename, '/');
+			tmp = (tmp == NULL ? elf->filename : tmp+1);
 			xstrcat(&out_buffer, tmp, &out_len);
 			break;
 		case 'o': out = get_elfetype(elf); break;
@@ -1005,14 +1009,115 @@ static void scanelf_file(const char *filename)
 
 	if (!found_file && (!be_quiet || (be_quiet && FOUND_SOMETHING()))) {
 		xchrcat(&out_buffer, ' ', &out_len);
-		xstrcat(&out_buffer, filename, &out_len);
+		xstrcat(&out_buffer, elf->filename, &out_len);
 	}
 	if (!be_quiet || (be_quiet && FOUND_SOMETHING())) {
 		puts(out_buffer);
 		fflush(stdout);
 	}
 
+	return 0;
+}
+
+/* scan a single elf */
+static int scanelf_elf(const char *filename, int fd, size_t len)
+{
+	int ret = 1;
+	elfobj *elf;
+
+	/* verify this is real ELF */
+	if ((elf = _readelf_fd(filename, fd, len, !fix_elf)) == NULL) {
+		if (be_verbose > 2) printf("%s: not an ELF\n", filename);
+		return ret;
+	}
+	switch (match_bits) {
+		case 32:
+			if (elf->elf_class != ELFCLASS32)
+				goto label_done;
+			break;
+		case 64:
+			if (elf->elf_class != ELFCLASS64)
+				goto label_done;
+			break;
+		default: break;
+	}
+	if (strlen(match_etypes)) {
+		char sbuf[126];
+		strncpy(sbuf, match_etypes, sizeof(sbuf));
+		if (strchr(match_etypes, ',') != NULL) {
+			char *p;
+			while((p = strrchr(sbuf, ',')) != NULL) {
+				*p = 0;
+				if (atoi(p+1) == get_etype(elf))
+					goto label_ret;
+			}
+		}
+		if (atoi(sbuf) != get_etype(elf))
+			goto label_done;
+	}
+
+label_ret:
+	ret = scanelf_elfobj(elf);
+
+label_done:
 	unreadelf(elf);
+	return ret;
+}
+
+/* scan an archive of elfs */
+static int scanelf_archive(const char *filename, int fd, size_t len)
+{
+	archive_handle *ar;
+	archive_member *m;
+	char *ar_buffer;
+	elfobj *elf;
+
+	ar = ar_open_fd(filename, fd);
+	if (ar == NULL)
+		return 1;
+
+	ar_buffer = (char*)mmap(0, len, PROT_READ | (fix_elf ? PROT_WRITE : 0), (fix_elf ? MAP_SHARED : MAP_PRIVATE), fd, 0);
+	while ((m=ar_next(ar)) != NULL) {
+		elf = readelf_buffer(m->name, ar_buffer+lseek(fd,0,SEEK_CUR), m->size);
+		if (elf) {
+			scanelf_elfobj(elf);
+			unreadelf(elf);
+		}
+	}
+	munmap(ar_buffer, len);
+
+	return 0;
+}
+/* scan a file which may be an elf or an archive or some other magical beast */
+static void scanelf_file(const char *filename)
+{
+	struct stat st;
+	int fd;
+
+	/* make sure 'filename' exists */
+	if (lstat(filename, &st) == -1) {
+		if (be_verbose > 2) printf("%s: does not exist\n", filename);
+		return;
+	}
+
+	/* always handle regular files and handle symlinked files if no -y */
+	if (S_ISLNK(st.st_mode)) {
+		if (!scan_symlink) return;
+		stat(filename, &st);
+	}
+	if (!S_ISREG(st.st_mode)) {
+		if (be_verbose > 2) printf("%s: skipping non-file\n", filename);
+		return;
+	}
+
+	if ((fd=open(filename, (fix_elf ? O_RDWR : O_RDONLY))) == -1)
+		return;
+
+	if (scanelf_elf(filename, fd, st.st_size) == 1 && scan_archives)
+		/* if it isn't an ELF, maybe it's an .a archive */
+		scanelf_archive(filename, fd, st.st_size);
+
+	close(fd);
 }
 
 /* scan a directory for ET_EXEC files and print when we find one */
@@ -1021,7 +1126,7 @@ static void scanelf_dir(const char *path)
 	register DIR *dir;
 	register struct dirent *dentry;
 	struct stat st_top, st;
-	char buf[_POSIX_PATH_MAX];
+	char buf[__PAX_UTILS_PATH_MAX];
 	size_t pathlen = 0, len = 0;
 
 	/* make sure path exists */
@@ -1041,7 +1146,7 @@ static void scanelf_dir(const char *path)
 		warnf("could not opendir %s: %s", path, strerror(errno));
 		return;
 	}
-	if (be_verbose) printf("%s: scanning dir\n", path);
+	if (be_verbose > 1) printf("%s: scanning dir\n", path);
 
 	pathlen = strlen(path);
 	while ((dentry = readdir(dir))) {
@@ -1070,14 +1175,14 @@ static int scanelf_from_file(char *filename)
 {
 	FILE *fp = NULL;
 	char *p;
-	char path[_POSIX_PATH_MAX];
+	char path[__PAX_UTILS_PATH_MAX];
 
 	if (((strcmp(filename, "-")) == 0) && (ttyname(0) == NULL))
 		fp = stdin;
 	else if ((fp = fopen(filename, "r")) == NULL)
 		return 1;
 
-	while ((fgets(path, _POSIX_PATH_MAX, fp)) != NULL) {
+	while ((fgets(path, __PAX_UTILS_PATH_MAX, fp)) != NULL) {
 		if ((p = strchr(path, '\n')) != NULL)
 			*p = 0;
 		search_path = path;
@@ -1092,13 +1197,13 @@ static void load_ld_so_conf()
 {
 	FILE *fp = NULL;
 	char *p;
-	char path[_POSIX_PATH_MAX];
+	char path[__PAX_UTILS_PATH_MAX];
 	int i = 0;
 
 	if ((fp = fopen("/etc/ld.so.conf", "r")) == NULL)
 		return;
 
-	while ((fgets(path, _POSIX_PATH_MAX, fp)) != NULL) {
+	while ((fgets(path, __PAX_UTILS_PATH_MAX, fp)) != NULL) {
 		if (*path != '/')
 			continue;
 
@@ -1161,7 +1266,7 @@ static void scanelf_envpath()
 
 
 /* usage / invocation handling functions */
-#define PARSE_FLAGS "plRmyXxetrnLibSs:gN:TaqvF:f:o:BhV"
+#define PARSE_FLAGS "plRmyAXxetrnLibSs:gN:TaqvF:f:o:M:E:BhV"
 #define a_argument required_argument
 static struct option const long_opts[] = {
 	{"path",      no_argument, NULL, 'p'},
@@ -1169,13 +1274,14 @@ static struct option const long_opts[] = {
 	{"recursive", no_argument, NULL, 'R'},
 	{"mount",     no_argument, NULL, 'm'},
 	{"symlink",   no_argument, NULL, 'y'},
+	{"archives",  no_argument, NULL, 'A'},
+	{"ldcache",   no_argument, NULL, 'L'},
 	{"fix",       no_argument, NULL, 'X'},
 	{"pax",       no_argument, NULL, 'x'},
 	{"header",    no_argument, NULL, 'e'},
 	{"textrel",   no_argument, NULL, 't'},
 	{"rpath",     no_argument, NULL, 'r'},
 	{"needed",    no_argument, NULL, 'n'},
-	{"ldcache",   no_argument, NULL, 'L'},
 	{"interp",    no_argument, NULL, 'i'},
 	{"bind",      no_argument, NULL, 'b'},
 	{"soname",    no_argument, NULL, 'S'},
@@ -1183,6 +1289,8 @@ static struct option const long_opts[] = {
 	{"lib",        a_argument, NULL, 'N'},
 	{"gmatch",    no_argument, NULL, 'g'},
 	{"textrels",  no_argument, NULL, 'T'},
+	{"etype",      a_argument, NULL, 'E'},
+	{"bits",       a_argument, NULL, 'M'},
 	{"all",       no_argument, NULL, 'a'},
 	{"quiet",     no_argument, NULL, 'q'},
 	{"verbose",   no_argument, NULL, 'v'},
@@ -1201,13 +1309,14 @@ static const char *opts_help[] = {
 	"Scan directories recursively",
 	"Don't recursively cross mount points",
 	"Don't scan symlinks",
+	"Scan archives (.a files)",
+	"Utilize ld.so.cache information (use with -r/-n)",
 	"Try and 'fix' bad things (use with -r/-e)\n",
 	"Print PaX markings",
 	"Print GNU_STACK/PT_LOAD markings",
 	"Print TEXTREL information",
 	"Print RPATH information",
 	"Print NEEDED information",
-	"Utilize ld.so.cache information (use with -r/-n)",
 	"Print INTERP information",
 	"Print BIND information",
 	"Print SONAME information",
@@ -1215,6 +1324,8 @@ static const char *opts_help[] = {
 	"Find a specified library",
 	"Use strncmp to match libraries. (use with -N)",
 	"Locate cause of TEXTREL",
+	"Print only ELF files matching numeric constant type",
+	"Print only ELF files matching numeric bits",
 	"Print all scanned info (-x -e -t -r -b)\n",
 	"Only output 'bad' things",
 	"Be verbose (can be specified more than once)",
@@ -1232,7 +1343,7 @@ static void usage(int status)
 {
 	unsigned long i;
 	printf("* Scan ELF binaries for stuff\n\n"
-	       "Usage: %s [options] <dir1/file1> [dir2 dirN fileN ...]\n\n", argv0);
+	       "Usage: %s [options] <dir1/file1> [dir2 dirN file2 fileN ...]\n\n", argv0);
 	printf("Options: -[%s]\n", PARSE_FLAGS);
 	for (i = 0; long_opts[i].name; ++i)
 		if (long_opts[i].has_arg == no_argument)
@@ -1255,6 +1366,9 @@ static void usage(int status)
 	puts(" f filename (short name/basename)");
 	puts("Prefix each modifier with '%' (verbose) or '#' (silent)");
 
+	puts("\nELF Etypes:");
+	print_etypes(stdout);
+
 	exit(status);
 }
 
@@ -1279,6 +1393,12 @@ static void parseargs(int argc, char *argv[])
 			if (from_file) warn("You prob don't want to specify -f twice");
 			from_file = optarg;
 			break;
+		case 'E':
+			strncpy(match_etypes, optarg, sizeof(match_etypes));
+			break;
+		case 'M':
+			match_bits = atoi(optarg);
+			break;
 		case 'o': {
 			FILE *fp = NULL;
 			if ((fp = freopen(optarg, "w", stdout)) == NULL)
@@ -1309,6 +1429,7 @@ static void parseargs(int argc, char *argv[])
 		case 'g': gmatch = 1; /* break; any reason we dont breal; here ? */
 		case 'L': use_ldcache = 1; break;
 		case 'y': scan_symlink = 0; break;
+		case 'A': scan_archives = 1; break;
 		case 'B': show_banner = 0; break;
 		case 'l': scan_ldpath = 1; break;
 		case 'p': scan_envpath = 1; break;
@@ -1329,11 +1450,11 @@ static void parseargs(int argc, char *argv[])
 		case 'a': show_pax = show_phdr = show_textrel = show_rpath = show_bind = 1; break;
 
 		case ':':
-			err("Option missing parameter\n");
+			err("Option '%c' is missing parameter", optopt);
 		case '?':
-			err("Unknown option\n");
+			err("Unknown option '%c' or argument missing", optopt);
 		default:
-			err("Unhandled option '%c'", i);
+			err("Unhandled option '%c'; please report this", i);
 		}
 	}
 
